"""

    KeepNote
    Application Options Dialog

"""

#
#  KeepNote
#  Copyright (c) 2008-2009 Matt Rasmussen
#  Author: Matt Rasmussen <rasmus@mit.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#

# python imports
import os
import sys
import gettext

# pygtk imports
import pygtk
pygtk.require('2.0')
import gtk.glade
from gtk import gdk

# keepnote imports
import keepnote
from keepnote import unicode_gtk
from keepnote import get_resource
from keepnote.gui.font_selector import FontSelector
import keepnote.gui
from keepnote.gui.icons import get_icon_filename
import keepnote.trans
import keepnote.gui.extension

_ = keepnote.translate




def on_browse(parent, title, filename, entry,
              action=gtk.FILE_CHOOSER_ACTION_OPEN):
    """Callback for selecting file browser associated with a text entry"""

    dialog = gtk.FileChooserDialog(title, parent, 
        action=action,
        buttons=(_("Cancel"), gtk.RESPONSE_CANCEL,
                 _("Open"), gtk.RESPONSE_OK))
    dialog.set_transient_for(parent)
    dialog.set_modal(True)

    # set the filename if it is fully specified
    if filename == "":
        filename = entry.get_text()
    if os.path.isabs(filename):            
        dialog.set_filename(filename)

    if dialog.run() == gtk.RESPONSE_OK and dialog.get_filename():
        entry.set_text(dialog.get_filename())

    dialog.destroy()


class Section (object):
    """A Section in the Options Dialog"""

    def __init__(self, key, dialog, app, label=u"", icon=None):
        self.key = key
        self.dialog = dialog
        self.label = label
        self.icon = icon

        self.frame = gtk.Frame("")
        self.frame.get_label_widget().set_text("<b>%s</b>" % label)
        self.frame.get_label_widget().set_use_markup(True)
        self.frame.set_property("shadow-type", gtk.SHADOW_NONE)

        self.__align = gtk.Alignment()
        self.__align.set_padding(10, 0, 10, 0)
        self.__align.show()
        self.frame.add(self.__align)


    def get_default_widget(self):
        """Returns the default parent widget for a Section"""
        return self.__align


    def load_options(self, app):
        """Load options from app to UI"""
        pass

    def save_options(self, app):
        """Save options to the app"""
        pass


class GeneralSection (Section):

    def __init__(self, key, dialog, app,
                 label=u"", icon="keepnote-16x16.png"):
        Section.__init__(self, key, dialog, app, label, icon)
        
        self.notebook = None
        
        self.xml = gtk.glade.XML(get_resource("rc", "keepnote.glade"),
                                 "general_frame", keepnote.GETTEXT_DOMAIN)
        self.xml.signal_autoconnect(self)
        self.xml.signal_autoconnect({
            "on_default_notebook_button_clicked":
                lambda w: on_browse(self.dialog,
                    _("Choose Default Notebook"),
                    "",
                    self.xml.get_widget("default_notebook_entry")),
            })
        self.frame = self.xml.get_widget("general_frame")



    def on_default_notebook_radio_changed(self, radio):
        """Default notebook radio changed"""
        no_default = self.xml.get_widget("no_default_notebook_radio")
        default = self.xml.get_widget("default_notebook_radio")
        last = self.xml.get_widget("last_notebook_radio")

        default_tab = self.xml.get_widget("default_notebook_table")
        default_tab.set_sensitive(default.get_active())
            

    def on_autosave_check_toggled(self, widget):
        """The autosave option controls sensitivity of autosave time"""
        self.xml.get_widget("autosave_entry").set_sensitive(
            widget.get_active())
        self.xml.get_widget("autosave_label").set_sensitive(
            widget.get_active())


    def on_systray_check_toggled(self, widget):
        """Systray option controls sensitivity of skip taskbar"""
        self.xml.get_widget("skip_taskbar_check").set_sensitive(
            widget.get_active())
        
    
    def on_set_default_notebook_button_clicked(self, widget):

        if self.notebook:
            self.xml.get_widget("default_notebook_entry").set_text(
                self.notebook.get_path())


    def load_options(self, app):
        
        win = app.get_current_window()
        if win:
            self.notebook = win.get_notebook()

        # populate default notebook        
        if app.pref.use_last_notebook:
            self.xml.get_widget("last_notebook_radio").set_active(True)
        elif app.pref.default_notebook == "":
            self.xml.get_widget("no_default_notebook_radio").set_active(True)
        else:
            self.xml.get_widget("default_notebook_radio").set_active(True)
            self.xml.get_widget("default_notebook_entry").\
                set_text(app.pref.default_notebook)



        # populate autosave
        self.xml.get_widget("autosave_check").set_active(
            app.pref.autosave)
        self.xml.get_widget("autosave_entry").set_text(
            str(int(app.pref.autosave_time / 1000)))

        self.xml.get_widget("autosave_entry").set_sensitive(
            app.pref.autosave)
        self.xml.get_widget("autosave_label").set_sensitive(
            app.pref.autosave)


        # use systray icon
        self.xml.get_widget("systray_check").set_active(app.pref.use_systray)
        self.xml.get_widget("skip_taskbar_check").set_active(app.pref.skip_taskbar)
        self.xml.get_widget("skip_taskbar_check").set_sensitive(app.pref.use_systray)


    def save_options(self, app):
        if self.xml.get_widget("last_notebook_radio").get_active():
            app.pref.use_last_notebook = True
        elif self.xml.get_widget("default_notebook_radio").get_active():
            app.pref.use_last_notebook = False
            app.pref.default_notebook = \
                unicode_gtk(self.xml.get_widget("default_notebook_entry").get_text())
        else:
            app.pref.use_last_notebook = False
            app.pref.default_notebook = ""


        # save autosave
        app.pref.autosave = \
            self.xml.get_widget("autosave_check").get_active()
        try:
            app.pref.autosave_time = \
                int(self.xml.get_widget("autosave_entry").get_text()) * 1000
        except:
            pass

        # use systray icon
        app.pref.use_systray = self.xml.get_widget("systray_check").get_active()
        app.pref.skip_taskbar = self.xml.get_widget("skip_taskbar_check").get_active()


class LookAndFeelSection (Section):
    
    def __init__(self, key, dialog, app, label=u"", icon="lookandfeel.png"):
        Section.__init__(self, key, dialog, app, label, icon)

        w = self.get_default_widget()
        v = gtk.VBox(False, 5)
        v.show()
        w.add(v)

        def add_check(label):
            c = gtk.CheckButton(label)
            c.show()
            v.pack_start(c, False, False, 0)
            return c

        self.treeview_lines_check = add_check(
            _("show lines in treeview"))
        self.listview_rules_check = add_check(
            _("use ruler hints in listview"))
        self.use_stock_icons_check = add_check(
            _("use GTK stock icons in toolbar"))
        self.use_minitoolbar = add_check(
            _("use minimal toolbar"))
        
        # view mode combo
        h = gtk.HBox(False, 5); h.show()
        l = gtk.Label(_("Listview Layout:")); l.show()
        h.pack_start(l, False, False, 0)
        c = gtk.combo_box_new_text(); c.show()
        c.append_text(_("Vertical"))
        c.append_text(_("Horizontal"))
        h.pack_start(c, False, False, 0)
        v.pack_start(h)
        self.listview_layout = c
        


    def load_options(self, app):

        self.treeview_lines_check.set_active(app.pref.treeview_lines)
        self.listview_rules_check.set_active(app.pref.listview_rules)
        self.use_stock_icons_check.set_active(app.pref.use_stock_icons)
        self.use_minitoolbar.set_active(app.pref.use_minitoolbar)

        if app.pref.view_mode == "horizontal":
            self.listview_layout.set_active(1)
        else:
            self.listview_layout.set_active(0)
            


    def save_options(self, app):
        
        app.pref.treeview_lines = self.treeview_lines_check.get_active()
        app.pref.listview_rules = self.listview_rules_check.get_active()
        app.pref.use_stock_icons = self.use_stock_icons_check.get_active()
        app.pref.use_minitoolbar = self.use_minitoolbar.get_active()

        app.pref.view_mode = ["vertical", "horizontal"][
            self.listview_layout.get_active()]
 

class LanguageSection (Section):
    
    def __init__(self, key, dialog, app, label=u"", icon=None):
        Section.__init__(self, key, dialog, app, label, icon)

        w = self.get_default_widget()
        v = gtk.VBox(False, 5)
        v.show()
        w.add(v)        
        
        # language combo
        h = gtk.HBox(False, 5); h.show()
        l = gtk.Label(_("Language:")); l.show()
        h.pack_start(l, False, False, 0)
        c = gtk.combo_box_new_text(); c.show()

        # populate language options
        c.append_text("default")
        for lang in keepnote.trans.get_langs():
            c.append_text(lang)
        
        # pack combo
        h.pack_start(c, False, False, 0)
        v.pack_start(h)
        self.language_box = c
        


    def load_options(self, app):

        # set default
        if app.pref.language == "":
            self.language_box.set_active(0)
        else:
            for i, row in enumerate(self.language_box.get_model()):
                if app.pref.language == row[0]:
                    self.language_box.set_active(i)
                    break


    def save_options(self, app):
        
        if self.language_box.get_active() > 0:
            app.pref.language = lang = self.language_box.get_active_text()
        else:
            # set default
            app.pref.language = ""

        # XXX: may be I should not change translation during execution
        #if app.pref.language != keepnote.trans.get_lang():
        #    keepnote.trans.set_lang(app.pref.language)
 



class HelperAppsSection (Section):
    
    def __init__(self, key, dialog, app, label=u"", icon=None):
        Section.__init__(self, key, dialog, app, label, icon)
        
        self.entries = {}
        w = self.get_default_widget()
        
        self.table = gtk.Table(len(app.pref.external_apps), 2)
        self.table.show()
        w.add(self.table)

        # set icon
        try:
            self.icon = keepnote.gui.get_pixbuf(get_icon_filename(gtk.STOCK_EXECUTE))
        except:
            pass


        
    def load_options(self, app):

        # clear table, resize
        self.table.foreach(lambda x: self.table.remove(x))
        self.table.resize(len(app.pref.external_apps), 2)

        for i, app in enumerate(app.pref.external_apps):
            key = app.key
            app_title = app.title
            prog = app.prog
            
            # program label
            label = gtk.Label(app_title +":")
            label.set_justify(gtk.JUSTIFY_RIGHT)
            label.set_alignment(1.0, 0.5)
            label.show()
            self.table.attach(label, 0, 1, i, i+1,
                         xoptions=gtk.FILL, yoptions=0,
                         xpadding=2, ypadding=2)

            # program entry
            entry = gtk.Entry()
            entry.set_text(prog)
            entry.set_width_chars(30)
            entry.show()
            self.entries[key] = entry
            self.table.attach(entry, 1, 2, i, i+1,
                         xoptions=gtk.FILL | gtk.EXPAND, yoptions=0,
                         xpadding=2, ypadding=2)

            # browse button
            def button_clicked(key, title, prog):
                return lambda w: \
                    on_browse(self.dialog,
                              _("Choose %s") % title,
                              "", self.entries[key])
            button = gtk.Button(_("Browse..."))
            button.set_image(
                gtk.image_new_from_stock(gtk.STOCK_OPEN,
                                         gtk.ICON_SIZE_SMALL_TOOLBAR))
            button.show()
            button.connect("clicked", button_clicked(key, app_title, prog))
            self.table.attach(button, 2, 3, i, i+1,
                         xoptions=0, yoptions=0,
                         xpadding=2, ypadding=2)
    

    def save_options(self, app):

        # TODO: use a public interface

        # save external app options
        for key, entry in self.entries.iteritems():
            if key in app.pref._external_apps_lookup:
                app.pref._external_apps_lookup[key].prog = unicode_gtk(
                    entry.get_text())




class DatesSection (Section):
    
    def __init__(self, key, dialog, app, label=u"", icon="time.png"):
        Section.__init__(self, key, dialog, app, label, icon)

        self.date_xml = gtk.glade.XML(get_resource("rc", "keepnote.glade"),
                                   "date_time_frame", keepnote.GETTEXT_DOMAIN)
        self.date_xml.signal_autoconnect(self)
        self.frame = self.date_xml.get_widget("date_time_frame")


    def load_options(self, app):
        for name in ["same_day", "same_month", "same_year", "diff_year"]:
            self.date_xml.get_widget("date_%s_entry" % name).\
                set_text(app.pref.timestamp_formats[name])

    def save_options(self, app):
        # save date formatting
        for name in ["same_day", "same_month", "same_year", "diff_year"]:
            app.pref.timestamp_formats[name] = unicode_gtk(
                self.date_xml.get_widget("date_%s_entry" % name).get_text())
        

class AllNoteBooksSection (Section):
    
    def __init__(self, key, dialog, app, label=u"", icon="folder.png"):
        Section.__init__(self, key, dialog, app, label, icon)

        w = self.get_default_widget()
        l = gtk.Label(_("This section contains options that are saved on a per notebook basis (e.g. notebook-specific font).   A subsection will appear for each notebook that is currently opened."))
        l.set_line_wrap(True)
        w.add(l)
        w.show_all()


class NoteBookSection (Section):
    
    def __init__(self, key, dialog, app, notebook, label=u"", 
                 icon="folder.png"):
        Section.__init__(self, key, dialog, app, label, icon)
        self.entries = {}

        self.notebook = notebook

        # add notebook font widget
        self.notebook_xml = gtk.glade.XML(get_resource("rc", "keepnote.glade"),
                                 "notebook_frame", keepnote.GETTEXT_DOMAIN)
        self.notebook_xml.signal_autoconnect(self)
        self.frame = self.notebook_xml.get_widget("notebook_frame")

        notebook_font_spot = self.notebook_xml.get_widget("notebook_font_spot")
        self.notebook_font_family = FontSelector()
        notebook_font_spot.add(self.notebook_font_family)
        self.notebook_font_family.show()        

        # populate notebook font
        self.notebook_font_size = self.notebook_xml.get_widget("notebook_font_size")
        self.notebook_font_size.set_value(10)
        self.notebook_index_dir = self.notebook_xml.get_widget("index_dir_entry")
        self.notebook_xml.get_widget("index_dir_browse").connect(
            "clicked",
            lambda w: on_browse(self.dialog,
                _("Choose alternative notebook index directory"),
                "", self.notebook_index_dir,
                action=gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER))

        self.frame.show_all()



    def load_options(self, app):
        
        if self.notebook is not None:
            font = self.notebook.pref.default_font
            family, mods, size = keepnote.gui.richtext.parse_font(font)
            self.notebook_font_family.set_family(family)
            self.notebook_font_size.set_value(size)

            self.notebook_index_dir.set_text(self.notebook.pref.index_dir)


    def save_options(self, app):
        
        if self.notebook is not None:
            pref = self.notebook.pref

            # save notebook font        
            pref.default_font = "%s %d" % (
                self.notebook_font_family.get_family(),
                self.notebook_font_size.get_value())

            # alternative index directory
            pref.index_dir = self.notebook_index_dir.get_text()

        

class ExtensionsSection (Section):
    
    def __init__(self, key, dialog, app, label=u"", icon=None):
        Section.__init__(self, key, dialog, app, label, icon)
        
        self.app = app
        self.entries = {}
        self.frame = gtk.Frame("")
        self.frame.get_label_widget().set_text("<b>Extensions</b>")
        self.frame.get_label_widget().set_use_markup(True)
        self.frame.set_property("shadow-type", gtk.SHADOW_NONE)
        
        align = gtk.Alignment()
        align.set_padding(10, 0, 10, 0)
        align.show()
        self.frame.add(align)
        
        v = gtk.VBox(False, 0)
        v.show()
        align.add(v)

        # extension list scrollbar
        self.sw = gtk.ScrolledWindow()
        self.sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.sw.set_shadow_type(gtk.SHADOW_IN)
        self.sw.show()
        v.pack_start(self.sw, True, True, 0)
        
        # extension list
        self.extlist = gtk.VBox(False, 0)
        self.extlist.show()
        self.sw.add_with_viewport(self.extlist)

        # hbox
        h = gtk.HBox(False, 0)
        h.show()
        v.pack_start(h, True, True, 0)
        
        # install button
        self.install_button = gtk.Button("Install new extension")
        self.install_button.set_relief(gtk.RELIEF_NONE)
        self.install_button.modify_fg(gtk.STATE_NORMAL, gtk.gdk.Color(0, 0, 65535))
        self.install_button.connect("clicked", self._on_install)
        self.install_button.show()
        h.pack_start(self.install_button, False, True, 0)

        # set icon
        try:
            self.icon = keepnote.gui.get_pixbuf(get_icon_filename(gtk.STOCK_ADD))
        except:
            pass



    def load_options(self, app):

        # clear extension list
        self.extlist.foreach(self.extlist.remove)
        
        def callback(ext):
            return lambda w: self._on_uninstall(ext)

        # populate extension list
        exts = list(app.iter_extensions())
        d = {"user": 0, "system": 1}
        exts.sort(key=lambda e: (d.get(e.type, 10), e.name))
        for ext in exts:
            if ext.visible:
                p = ExtensionWidget(app, ext)
                p.uninstall_button.connect("clicked", callback(ext))
                p.show()
                self.extlist.pack_start(p, True, True, 0)
        
        # setup scroll bar size
        maxheight = 270 # TODO: make this more dynamic
        w, h = self.extlist.size_request()
        w2, h2 = self.sw.get_vscrollbar().size_request()
        self.sw.set_size_request(400, min(maxheight, h+10))


    def save_options(self, app):
        
        for widget in self.extlist:
            try:
                widget.ext.enable(widget.enabled)
            except Exception, e:
                keepnote.log_error(e)
            widget.update()


    def _on_uninstall(self, ext):
        if self.app.uninstall_extension(ext):
            self.load_options(self.app)


    def _on_install(self, widget):

        # open file dialog
        dialog = gtk.FileChooserDialog(
            _("Install New Extension"), self.dialog, 
            action=gtk.FILE_CHOOSER_ACTION_OPEN,
            buttons=(_("Cancel"), gtk.RESPONSE_CANCEL,
                     _("Open"), gtk.RESPONSE_OK))
        dialog.set_transient_for(self.dialog)
        dialog.set_modal(True)

        file_filter = gtk.FileFilter()
        file_filter.add_pattern("*.kne")
        file_filter.set_name(_("KeepNote Extension (*.kne)"))
        dialog.add_filter(file_filter)
        
        file_filter = gtk.FileFilter()
        file_filter.add_pattern("*")
        file_filter.set_name(_("All files (*.*)"))
        dialog.add_filter(file_filter)

        response = dialog.run()

        if gtk.RESPONSE_OK and dialog.get_filename():
            # install extension
            self.app.install_extension(dialog.get_filename())
            self.load_options(self.app)

        dialog.destroy()



class ExtensionWidget (gtk.EventBox):
    def __init__(self, app, ext):
        gtk.EventBox.__init__(self)

        self.app = app
        self.enabled = ext.is_enabled()
        self.ext = ext

        self.modify_bg(gtk.STATE_NORMAL, gtk.gdk.Color(65535, 65535, 65535))

        frame = gtk.Frame(None)
        frame.set_property("shadow-type", gtk.SHADOW_OUT)
        frame.show()
        self.add(frame)

        # name
        frame2 = gtk.Frame("")
        frame2.set_property("shadow-type", gtk.SHADOW_NONE)
        frame2.get_label_widget().set_text("<b>%s</b> (%s/%s)" % 
                                           (ext.name, ext.type,
                                            ext.key))
        frame2.get_label_widget().set_use_markup(True)
        frame2.show()
        frame.add(frame2)

        # margin
        align = gtk.Alignment()
        align.set_padding(10, 10, 10, 10)
        align.show()
        frame2.add(align)

        # vbox
        v = gtk.VBox(False, 5)
        v.show()
        align.add(v)

        # description
        l = gtk.Label(ext.description)
        l.set_justify(gtk.JUSTIFY_LEFT)
        l.set_alignment(0.0, 0.0)
        l.show()
        v.pack_start(l, True, True, 0)

        # hbox
        h = gtk.HBox(False, 0)
        h.show()
        v.pack_start(h, True, True, 0)

        # enable button
        self.enable_check = gtk.CheckButton("Enabled")
        self.enable_check.set_active(self.enabled)
        self.enable_check.show()
        self.enable_check.connect(
            "toggled", lambda w: self._on_enabled(ext))
        h.pack_start(self.enable_check, False, True, 0)

        # divider
        l = gtk.Label("|"); l.show()
        h.pack_start(l, False, True, 0)

        # uninstall button        
        self.uninstall_button = gtk.Button("Uninstall")
        self.uninstall_button.set_relief(gtk.RELIEF_NONE)
        self.uninstall_button.set_sensitive(app.can_uninstall(ext))
        self.uninstall_button.show()
        h.pack_start(self.uninstall_button, False, True, 0)



    def update(self):
        self.enable_check.set_active(self.ext.is_enabled())

    def _on_enabled(self, ext):
        self.enabled = self.enable_check.get_active()


        


#=============================================================================

class ApplicationOptionsDialog (object):
    """Application options"""
    
    def __init__(self, app):
        self.app = app
        self.parent = None

        self._sections = []
        self.tree2section = {}
        
        self.xml = gtk.glade.XML(get_resource("rc", "keepnote.glade"),
                                 "app_options_dialog", keepnote.GETTEXT_DOMAIN)
        self.dialog = self.xml.get_widget("app_options_dialog")
        self.dialog.connect("delete-event", self._on_delete_event)
        self.tabs = self.xml.get_widget("app_options_tabs")
        self.xml.signal_autoconnect({
            "on_cancel_button_clicked": 
                lambda w: self.on_cancel_button_clicked(),
            "on_ok_button_clicked":
                lambda w: self.on_ok_button_clicked(),
            "on_apply_button_clicked":
                lambda w: self.on_apply_button_clicked()})

        # setup treeview
        self.overview = self.xml.get_widget("app_config_treeview")
        self.overview_store = gtk.TreeStore(str, object, gdk.Pixbuf)
        self.overview.set_model(self.overview_store)
        self.overview.connect("cursor-changed", self.on_overview_select)

        # create the treeview column
        column = gtk.TreeViewColumn()
        self.overview.append_column(column)
        cell_text = gtk.CellRendererText()
        cell_icon = gtk.CellRendererPixbuf()
        column.pack_start(cell_icon, True)
        column.add_attribute(cell_icon, 'pixbuf', 2)
        column.pack_start(cell_text, True)
        column.add_attribute(cell_text, 'text', 0)
        

        # add tabs
        self.add_default_sections()

    
    def show(self, parent):
        """Display application options"""
        
        self.parent = parent
        self.dialog.set_transient_for(parent)

        # add notebook options
        self.notebook_sections = [
            self.add_section(NoteBookSection("notebook_%d" % i, 
                                             self.dialog, self.app, 
                                             notebook,
                                             notebook.get_title()), 
                             "notebooks")
            for i, notebook in enumerate(self.app.iter_notebooks())]


        # add extension options
        self.extensions_ui = []
        for ext in self.app.iter_extensions(True):
            if isinstance(ext, keepnote.gui.extension.Extension):
                ext.on_add_options_ui(self)
                self.extensions_ui.append(ext)

        # populate options ui
        self.load_options(self.app)

        self.dialog.show()

    def finish(self):

        # remove extension options
        for ext in self.extensions_ui:
            if isinstance(ext, keepnote.gui.extension.Extension):
                ext.on_remove_options_ui(self)
        self.extensions_ui = []

        # remove notebook options
        for section in self.notebook_sections:
            self.remove_section(section.key)


    def add_default_sections(self):

        self.add_section(
            GeneralSection("general", self.dialog, self.app, 
                           keepnote.PROGRAM_NAME))
        self.add_section(
            LookAndFeelSection("look_and_feel", self.dialog, 
                               self.app, _("Look and Feel")), 
            "general")
        self.add_section(
            LanguageSection("language", self.dialog, self.app, 
                            _("Language")),
            "general")
        self.add_section(
            DatesSection("date_and_time", self.dialog, self.app, 
                         _("Date and Time")), 
            "general")
        self.add_section(
            HelperAppsSection("helper_apps", self.dialog, 
                              self.app, _("Helper Applications")), 
            "general")
        self.add_section(
            AllNoteBooksSection("notebooks", self.dialog, self.app, 
                                _("Notebook Options"), "folder.png"))
        self.add_section(
            ExtensionsSection("extensions", self.dialog, 
                              self.app, _("Extensions")))

    #=====================================
    # options

    def load_options(self, app):
        """Load options into sections"""
        
        for section in self._sections:
            section.load_options(self.app)

    
    def save_options(self, app):
        """Save the options from each section"""
        
        for section in self._sections:
            section.save_options(self.app)

        # notify application preference changes
        self.app.pref.changed.notify()

        # save noteboook preference changes
        for notebook in self.app.iter_notebooks():
            notebook.write_preferences()
            notebook.notify_change(False)


    #=====================================
    # section handling

    def add_section(self, section, parent=None):
        """Add a section to the Options Dialog"""
        
        # determine parent section
        if parent is not None:
            path = self.get_section_path(parent)
            it = self.overview_store.get_iter(path)
        else:
            it = None

        self._sections.append(section)
        self.tabs.insert_page(section.frame, tab_label=None)
        section.frame.show()
        section.frame.queue_resize()

        icon = section.icon
        if icon is None:
            icon = icon="note.png"
        
        if isinstance(icon, basestring):
            pixbuf = keepnote.gui.get_resource_pixbuf(icon)
        else:
            pixbuf = icon

        # add to overview
        it = self.overview_store.append(it, [section.label, section, pixbuf])
        path = self.overview_store.get_path(it)
        self.overview.expand_to_path(path)
        self.tree2section[path] = section

        return section


    def remove_section(self, key):
        
        # TODO: may need to update tree2section, when other pages slide in position

        path = self.get_section_path(key)

        if path is None:
            return

        # remove from tabs
        section = self.tree2section[path]
        self.tabs.remove_page(self._sections.index(section))
        del self.tree2section[path]

        # remove from tree
        self.overview_store.remove(self.overview_store.get_iter(path))

        self._sections.remove(section)


    def get_section(self, key):
        """Returns the section for a key"""

        for section in self._sections:
            if section.key == key:
                return section

            
    def get_section_path(self, key):
        """Returns the TreeModel path for a section"""

        def walk(node):

            child = self.overview_store.iter_children(node)
            while child:
                row = self.overview_store[child]
                if row[1].key == key:
                    return row.path
                
                # recurse
                ret = walk(child)
                if ret:
                    return ret

                child = self.overview_store.iter_next(child)

            return None
        return walk(None)


    #==========================================================
    # callbacks

    def on_overview_select(self, overview):
        """Callback for changing topic in overview"""
        row, col = overview.get_cursor()
        if row is not None:
            section = self.tree2section[row]
            self.tabs.set_current_page(self._sections.index(section))


    def on_cancel_button_clicked(self):
        """Callback for cancel button"""
        self.dialog.hide()
        self.finish()

    
    def on_ok_button_clicked(self):
        """Callback for ok button"""
        self.save_options(self.app)
        self.dialog.hide()
        self.finish()


    def on_apply_button_clicked(self):
        """Callback for apply button"""
        self.save_options(self.app)

        # clean up and reshow dialog
        self.finish()
        self.show(self.parent)
        
    
    def _on_delete_event(self, widget, event):
        """Callback for window close"""
        self.dialog.hide()
        self.finish()
        self.dialog.stop_emission("delete-event")
        return True





#!/usr/bin/env python
# 
# setup for KeepNote
#
# use the following command to install KeepNote:
#   python setup.py install
#
#=============================================================================

#
#  KeepNote
#  Copyright (c) 2008-2009 Matt Rasmussen
#  Author: Matt Rasmussen <rasmus@mit.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#




#=============================================================================
# constants

import keepnote
KEEPNOTE_VERSION = keepnote.PROGRAM_VERSION_TEXT


#=============================================================================
# python and distutils imports
import os, sys, shutil, itertools
#from ez_setup import use_setuptools
#use_setuptools()
#from setuptools import setup, find_packages
from distutils.core import setup

# py2exe module (if building on windows)
try:
    import py2exe
except ImportError:
    pass


#=============================================================================
# helper functions

def split_path(path):
    """Splits a path into all of its directories"""

    pathlist = []
    while path != "":
        path, tail = os.path.split(path)
        pathlist.append(tail)
    pathlist.reverse()
    return pathlist
    


def get_files(path, exclude=lambda f: False):
    """Recursively get files from a directory"""
    files = []

    if isinstance(exclude, list):
        exclude_list = exclude
        def exclude(filename):
            for ext in exclude_list:
                if filename.endswith(ext):
                    return True
            return False

    def walk(path):
        for f in os.listdir(path):
            filename = os.path.join(path, f)
            if exclude(filename):
                # exclude certain files
                continue
            elif os.path.isdir(filename):
                # recurse directories                
                walk(filename)
            else:
                # record all other files
                files.append(filename)
    walk(path)

    return files

def get_file_lookup(files, prefix_old, prefix_new, 
                    exclude=lambda f: False):
    """Create a dictionary lookup of files"""

    if files is None:
        files = get_files(prefix_old, exclude=exclude)

    prefix_old = split_path(prefix_old)
    prefix_new = split_path(prefix_new)
    lookup = {}    

    for f in files:        
        path = prefix_new + split_path(f)[len(prefix_old):]
        dirpath = os.path.join(*path[:-1])
        lookup.setdefault(dirpath, []).append(f)

    return lookup


def remove_package_dir(filename):
    i = filename.index("/")
    return filename[i+1:]


#=============================================================================
# resource files/data

# get resources
rc_files = get_file_lookup(None, "keepnote/rc", "rc")
image_files = get_file_lookup(None, "keepnote/images", "images")
efiles = get_file_lookup(None, "keepnote/extensions", "extensions",
                         exclude=[".pyc"])
freedesktop_files = [
    # application icon
    ("share/icons/hicolor/48x48/apps",
     ["desktop/keepnote.png"]),

    # desktop menu entry
    ("share/applications",
     ["desktop/keepnote.desktop"])]


# get data files
if "py2exe" in sys.argv:
    data_files = rc_files.items() + efiles.items() + image_files.items()
    package_data = {}
    
else:
    data_files = freedesktop_files
    package_data = {'keepnote': []}
    for v in itertools.chain(rc_files.values(),
                             image_files.values(),
                             efiles.values()):
        package_data['keepnote'].extend(map(remove_package_dir, v))


#=============================================================================
# setup

setup(
    name='keepnote',
    version=KEEPNOTE_VERSION,
    description='A cross-platform note taking application',
    long_description = """
        KeepNote is a cross-platform note taking application.  Its features 
        include:
        
        - rich text editing
        
          - bullet points
          - fonts/colors
          - hyperlinks
          - inline images
          
        - hierarchical organization for notes
        - full text search
        - integrated screenshot
        - spell checking (via gtkspell)
        - backup and restore
        - HTML export
    """,
    author='Matt Rasmussen',
    author_email='rasmus@mit.edu',
    url='http://rasm.ods.org/keepnote/',
    download_url='http://rasm.ods.org/keepnote/download/keepnote-%s.tar.gz' % KEEPNOTE_VERSION,
    
    classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Environment :: Win32 (MS Windows)',
          'Environment :: X11 Applications',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: End Users/Desktop',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU General Public License (GPL)',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          ],
    license="GPL",
    
    packages=['keepnote',
              'keepnote.gui',
              'keepnote.gui.richtext',
              'keepnote.notebook',
              'keepnote.compat',
              'keepnote.mswin'],
    scripts=['bin/keepnote'],
    data_files=data_files,
    package_data=package_data,
    
    windows=[{
        'script': 'bin/keepnote',
        'icon_resources': [(1, 'keepnote/images/keepnote.ico')],
        }],
    options = {
        'py2exe' : {
            'packages': 'encodings',
            'includes': 'cairo,pango,pangocairo,atk,gobject,win32com.shell,win32api,win32com,win32ui,win32gui',
            'dist_dir': 'dist/keepnote-%s.win' % KEEPNOTE_VERSION
        },
        #'sdist': {
        #    'formats': 'zip',
        #}
    }
    )


# execute post-build script
if "py2exe" in sys.argv:
    execfile("pkg/win/post_py2exe.py")



"""

    KeepNote
    Notebook indexing

"""

#
#  KeepNote
#  Copyright (c) 2008-2009 Matt Rasmussen
#  Author: Matt Rasmussen <rasmus@mit.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#


# python imports
import os
import sys
import traceback
from thread import get_ident
import sqlite3  as sqlite
sqlite.enable_shared_cache(True)
#sqlite.threadsafety = 0

# keepnote imports
import keepnote


# index filename
INDEX_FILE = u"index.sqlite"
INDEX_VERSION = 1


def get_index_file(notebook):
    """Get the index filename for a notebook"""

    if notebook.pref.index_dir and os.path.exists(notebook.pref.index_dir):
        index_dir = notebook.pref.index_dir
    else:
        index_dir = notebook.get_pref_dir()

    return os.path.join(index_dir, INDEX_FILE)


def preorder(node):
    """Iterate through nodes in pre-order traversal"""

    queue = [node]

    while len(queue) > 0:
        node = queue.pop()
        yield node

        for child in node.iter_temp_children():
            queue.append(child)


class NoteBookIndexDummy (object):
    """Index for a NoteBook"""

    def __init__(self, notebook):
        pass

    def open(self):
        """Open connection to index"""
        pass

    def get_con(self):
        """Get connection for thread"""
        pass

    def close(self):
        """Close connection to index"""
        pass

    def init_index(self):
        """Initialize the tables in the index if they do not exist"""
        pass
            
    def add_node(self, node):
        """Add a node to the index"""
        pass

    def remove_node(self, node):
        """Remove node from index"""
        pass
        
    def get_node_path(self, nodeid):
        """Get node path for a nodeid"""
        return None

    def search_titles(self, query, cols=[]):
        """Return nodeids of nodes with matching titles"""
        return []

    def save(self):
        """Save index"""
        pass



class NoteBookIndex (object):
    """Index for a NoteBook"""

    def __init__(self, notebook):
        self._notebook = notebook
        self._uniroot = notebook.get_universal_root_id()
        self._need_index = False
        self._corrupt = False
        
        self.con = None
        self.cur = None
        self.open()

        self.add_node(notebook)


    def open(self):
        """
        Open connection to index
        """

        try:
            index_file = get_index_file(self._notebook)
            self._corrupt = False
            self.con = sqlite.connect(index_file, isolation_level="DEFERRED",
                                      check_same_thread=False)
            self.cur = self.con.cursor()
            self.con.execute(u"PRAGMA read_uncommitted = true;")

            self.init_index()
        except sqlite.DatabaseError, e:
            self._on_corrupt(e, sys.exc_info()[2])


    def close(self):
        """Close connection to index"""
        
        if self.con is not None:
            self.con.commit()
            self.con.close()
            self.con = None
            self.cur = None
    
    
    def is_corrupt(self):
        """Return True if database appear corrupt"""
        return self._corrupt


    def init_index(self):
        """Initialize the tables in the index if they do not exist"""

        self._need_index = False

        con = self.con

        try:

            # check database version
            con.execute(u"""CREATE TABLE IF NOT EXISTS Version 
                            (version INTEGER, update_date DATE);""")
            version = con.execute(u"SELECT MAX(version) FROM Version").fetchone()

            if version is None or version[0] != INDEX_VERSION:
                # version does not exist, drop all tables
                con.execute(u"DROP TABLE IF EXISTS NodeGraph")
                con.execute(u"DROP INDEX IF EXISTS IdxNodeGraphNodeid")
                con.execute(u"DROP INDEX IF EXISTS IdxNodeGraphParentid")
                con.execute(u"DROP TABLE IF EXISTS Nodes")
                con.execute(u"DROP TABLE IF EXISTS IdxNodesTitle")

                # update version
                con.execute(u"INSERT INTO Version VALUES (?, datetime('now'));", (INDEX_VERSION,))

                self._need_index = True


            # init NodeGraph table
            con.execute(u"""CREATE TABLE IF NOT EXISTS NodeGraph 
                           (nodeid TEXT,
                            parentid TEXT,
                            basename TEXT,
                            symlink BOOLEAN);
                        """)

            con.execute(u"""CREATE INDEX IF NOT EXISTS IdxNodeGraphNodeid 
                           ON NodeGraph (nodeid);""")
            con.execute(u"""CREATE INDEX IF NOT EXISTS IdxNodeGraphParentid 
                           ON NodeGraph (parentid);""")


            # init Nodes table
            # TODO: remove icon attribute
            # TODO: this is basically just a title table
            con.execute(u"""CREATE TABLE IF NOT EXISTS Nodes
                           (nodeid TEXT,
                            title TEXT,
                            icon TEXT);
                        """)

            con.execute(u"""CREATE INDEX IF NOT EXISTS IdxNodesTitle 
                           ON Nodes (Title);""")


            # TODO: make an Attr table
            # this will let me query whether an attribute is currently being
            # indexed and in what table it is in.
            #con.execute(u"""CREATE TABLE IF NOT EXISTS Attr
            #               (attr TEXT,
            #                tablename TEXT);
            #            """)

            con.commit()
        except sqlite.DatabaseError, e:
            self._on_corrupt(e, sys.exc_info()[2])

    
    def index_needed(self):
        return self._need_index

    
    def clear(self):
        """Erases database file and reinitializes"""

        self.close()
        index_file = get_index_file(self._notebook)
        if os.path.exists(index_file):
            os.remove(index_file)
        self.open()


    def index_all(self, root=None):
        """
        Reindex all nodes under root

        This function returns an iterator which must be iterated to completion.
        """
        
        if root is None:
            root = self._notebook
        
        visit = set()
        queue = []

        # record nodes that change while indexing
        def changed_callback(nodes, recurse):
            for node in nodes:
                if node not in visit:
                    queue.append(node)
        self._notebook.node_changed.add(changed_callback)

        # perform indexing
        for node in preorder(root):
            self.add_node(node)
            visit.add(node)
            yield node

        # walk through nodes missed in original pass
        while len(queue) > 0:
            node = queue.pop()
            if node not in visit:
                for node2 in preorder(node):
                    self.add_node(node)
                    visit.add(node)
                    yield node

        # remove callback for notebook changes
        self._notebook.node_changed.remove(changed_callback)

        # record index complete
        self._need_index = False

            
    def add_node(self, node):
        """Add a node to the index"""               

        if self.con is None:
            return
        con, cur = self.con, self.cur

        try:

            # TODO: remove single parent assumption

            # get info
            nodeid = node.get_attr("nodeid")
            parent = node.get_parent()
            if parent:
                parentid = parent.get_attr("nodeid")
                basename = node.get_basename()
            else:
                parentid = self._uniroot
                basename = u""
            symlink = False
            title = node.get_title()

            #------------------
            # NodeGraph
            rows = list(cur.execute(u"SELECT parentid, basename "
                                    u"FROM NodeGraph "
                                    u"WHERE nodeid = ?", (nodeid,)))
            if rows:
                row = rows[0]
                if row[0] != parentid or row[1] != basename:
                    # record update
                    ret = cur.execute(u"UPDATE NodeGraph SET "
                                      u"nodeid=?, "
                                      u"parentid=?, "
                                      u"basename=?, "
                                      u"symlink=? "
                                      u"WHERE nodeid = ?",
                                      (nodeid, parentid, basename, 
                                       symlink, nodeid))
            else:
                # insert new row
                cur.execute(u"""
                    INSERT INTO NodeGraph VALUES 
                       (?, ?, ?, ?)""",
                (nodeid,
                 parentid,
                 basename,
                 symlink,
                 ))

            #-----------------
            # Nodes
            rows = list(cur.execute(u"SELECT title "
                                    u"FROM Nodes "
                                    u"WHERE nodeid = ?", (nodeid,)))
            if rows:
                row = rows[0]     

                if row[0] != title:
                    # record update
                    ret = cur.execute(u"UPDATE Nodes SET "
                                      u"nodeid=?, "
                                      u"title=? "
                                      u"WHERE nodeid = ?",
                                      (nodeid, title, nodeid))
            else:
                # insert new row
                cur.execute(u"""
                    INSERT INTO Nodes VALUES 
                       (?, ?, ?)""",
                (nodeid, title, u""))

        except sqlite.DatabaseError, e:
            self._on_corrupt(e, sys.exc_info()[2])


    def remove_node(self, node):
        """Remove node from index"""

        if self.con is None:
            return
        con, cur = self.con, self.cur

        try:
            # get info
            nodeid = node.get_attr("nodeid")

            # delete node
            cur.execute(
                u"DELETE FROM NodeGraph WHERE nodeid=?", (nodeid,))
            cur.execute(
                u"DELETE FROM Nodes WHERE nodeid=?", (nodeid,))
            #con.commit()

        except sqlite.DatabaseError, e:
            self._on_corrupt(e, sys.exc_info()[2])

        
    def get_node_path(self, nodeid):
        """Get node path for a nodeid"""
        
        # TODO: handle multiple parents

        con, cur = self.con, self.cur

        try:
            def walk(nodeid):
                cur.execute(u"""SELECT nodeid, parentid, basename
                                FROM NodeGraph
                                WHERE nodeid=?""", (nodeid,))
                row = cur.fetchone()

                if row:
                    nodeid, parentid, basename = row
                    if parentid != self._uniroot:
                        path = self.get_node_path(parentid)
                        if path is not None:
                            path.append(basename)
                            return path
                        else:
                            return None
                    else:
                        return [basename]
            return walk(nodeid)

        except sqlite.DatabaseError, e:
            self._on_corrupt(e, sys.exc_info()[2])


    def search_titles(self, query, cols=[]):
        """Return nodeids of nodes with matching titles"""

        con, cur = self.con, self.cur

        if cols:
            sql = u"," + u",".join(cols)
        else:
            sql = u""

        # order titles by exact matches and then alphabetically
        cur.execute(u"""SELECT nodeid, title %s FROM Nodes WHERE title LIKE ?
                       ORDER BY title != ?, title """ % sql,
                    (u"%" + query + u"%", query))
        
        return list(cur.fetchall())

        


    def save(self):
        """Save index"""

        if self.con is not None:
            self.con.commit()


    def _on_corrupt(self, error, tracebk=None):

        self._corrupt = True

        # display error
        keepnote.log_error(error, tracebk)


#NoteBookIndex = NoteBookIndexDummy





import sys

# make sure py2exe finds win32com
try:
    import modulefinder
    import win32com
    for p in win32com.__path__[1:]:
        modulefinder.AddPackagePath("win32com", p)
    for extra in ["win32com.shell"]:
        __import__(extra)
        m = sys.modules[extra]
        for p in m.__path__[1:]:
            modulefinder.AddPackagePath(extra, p)
except ImportError:
    # no build path setup, no worries.
    pass


try:
    from win32com.shell import shell
    import win32api
    import win32gui
    import win32con
    import win32ui

    import ctypes.windll.kernel32

except:
    pass


def get_my_documents():
    """Return the My Docuemnts folder"""
    
    df = shell.SHGetDesktopFolder()
    pidl = df.ParseDisplayName(0, None,  
        "::{450d8fba-ad25-11d0-98a8-0800361b1103}")[1]
    mydocs = shell.SHGetPathFromIDList(pidl)

    # TODO: may need to handle window-specific encoding here.
    #encoding = locale.getdefaultlocale()[1]
    #if encoding is None:
    #    encoding = "utf-8"

    return mydocs


#def set_env(key, val):
#    ctypes.windll.kernel32.SetEnvironmentVariableW(key, val)




"""
 
    KeepNote
    RichText base classes for tags

"""


#
#  KeepNote
#  Copyright (c) 2008-2009 Matt Rasmussen
#  Author: Matt Rasmussen <rasmus@mit.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#

# pygtk imports
import pygtk
pygtk.require('2.0')
import gtk, gobject, pango
from gtk import gdk



#=============================================================================
# tags and tag table


class RichTextBaseTagTable (gtk.TextTagTable):
    """A tag table for a RichTextBuffer"""

    # Class Tags:
    # Class tags cannot overlap any other tag of the same class.
    # example: a piece of text cannot have two colors, two families,
    # two sizes, or two justifications.

    
    def __init__(self):
        gtk.TextTagTable.__init__(self)

        self._tag_classes = {}
        self._tag2class = {}
        self._expiring_tags = set()
        self._buffers = set()
        self._next_gc_size = 100
        self._gc_size_step = 100

    def add_textbuffer(self, buf):
        self._buffers.add(buf)

    def remove_textbuffer(self, buf):
        if buf in self._buffers:
            self._buffers.remove(buf)

    def remove(self, tag):
        gtk.TextTagTable.remove(self, tag)
        if tag in self._expiring_tags:
            self._expiring_tags.remove(tag)
        cls = self._tag2class[tag]
        del self._tag2class[tag]        
        cls.tags.remove(tag)

        
            

    def new_tag_class(self, class_name, class_type, exclusive=True):
        """Create a new RichTextTag class for RichTextBaseTagTable"""
        c = RichTextTagClass(class_name, class_type, exclusive)
        self._tag_classes[class_name] = c
        return c

    def get_tag_class(self, class_name):
        """Return the set of tags for a class"""
        return self._tag_classes[class_name]

    def get_tag_class_type(self, class_name):
        """Return the RichTextTag type for a class"""
        return self._tag_classes[class_name].class_type

    def tag_class_add(self, class_name, tag):
        """Add a tag to a tag class"""
        c = self._tag_classes[class_name]
        c.tags.add(tag)
        self.add(tag)
        self._tag2class[tag] = c
        return tag
        

    def get_class_of_tag(self, tag):
        """Returns the exclusive class of tag,
           or None if not an exclusive tag"""
        return self._tag2class.get(tag, None)
    

    def lookup(self, name):
        """Lookup any tag, create it if needed"""

        # test to see if name is directly in table
        #  modifications and justifications are directly stored
        tag = gtk.TextTagTable.lookup(self, name)        
        if tag:
            return tag

        # make tag from scratch
        for tag_class in self._tag_classes.itervalues():
            if tag_class.class_type.is_name(name):
                tag = tag_class.class_type.make_from_name(name)
                self.tag_class_add(tag_class.name, tag)
                
                if tag.expires():
                    self.gc()
                    self._expiring_tags.add(tag)

                return tag
        
        
        raise Exception("unknown tag '%s'" % name)


    def gc(self):
        """Garbage collect"""

        if self.get_size() > self._next_gc_size:

            #print "before", self.get_size()

            saved = set()

            # test to see if any expiring texttags have completely expired
            for buf in self._buffers:

                # scan buffer for all present tags
                it = buf.get_start_iter()
                o = it.get_offset()
                while True:                    
                    for tag in it.get_tags():
                        if tag in self._expiring_tags:
                            saved.add(tag)

                    if (not it.forward_to_tag_toggle(None) or
                        it.get_offset() == o):
                        break
                    o = it.get_offset()
                    
            
            # remove expired tags
            remove = []
            for tag in self._expiring_tags:
                if tag not in saved:
                    remove.append(tag)
            for tag in remove:
                self.remove(tag)

            self._next_gc_size = self.get_size() + self._gc_size_step

            #def func(x, data):
            #    print "tab", x.get_property("name")
            #self.foreach(func)

            #print "after", self.get_size()
    


class RichTextTagClass (object):
    """
    A class of tags that specify the same attribute


    Class tags cannot overlap any other tag of the same class.
    example: a piece of text cannot have two colors, two families,
    two sizes, or two justifications.

    """

    def __init__(self, name, class_type, exclusive=True):
        """
        name:        name of the class of tags (i.e. "family", "fg_color")
        class_type:  RichTextTag class for all tags in class
        exclusive:   bool for whether tags in class should be mutually exclusive
        """
        
        self.name = name
        self.tags = set()
        self.class_type = class_type
        self.exclusive = exclusive



class RichTextTag (gtk.TextTag):
    """A TextTag in a RichTextBuffer"""
    def __init__(self, name, **kargs):
        gtk.TextTag.__init__(self, name)
        self._count = 0
        
        for key, val in kargs.iteritems():
            self.set_property(key.replace("_", "-"), val)

    def expires(self):
        return False

    def can_be_current(self):
        return True

    def can_be_copied(self):
        return True

    def is_par_related(self):
        return False

    @classmethod
    def tag_name(cls):
        # NOT implemented
        raise Exception("Not implemented")

    @classmethod
    def get_value(cls, tag_name):
        # NOT implemented
        raise Exception("Not implemented")

    @classmethod
    def is_name(cls, tag_name):
        return False

    @classmethod
    def make_from_name(cls, tag_name):        
        return cls(cls.get_value(tag_name))



"""

    XmlObject

    This module allows concise definitions of XML file formats for python
    objects.

"""

#
#  KeepNote
#  Copyright (c) 2008-2009 Matt Rasmussen
#  Author: Matt Rasmussen <rasmus@mit.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#


# python imports
import sys
import codecs

# xml imports
import xml.parsers.expat
from xml.sax.saxutils import escape

# keepnote imports
from keepnote import safefile




class XmlError (StandardError):
    """Error for parsing XML"""
    pass


def bool2str(b):
    """Convert a bool into a string"""
    return str(int(b))

def str2bool(s):
    """Convert a string into a bool"""
    return bool(int(s))
    
def str_no_none(x):
    if x is None:
        return u""
    return x


class Tag (object):
    def __init__(self, name,
                 get=None,
                 set=None,
                 attr=None,
                 tags=[]):
        self.name = name                
        
        self._tag_list = list(tags)
        self._read_data = get
        self._write_data = set
        self._object = None
        self._data = []

        # set read/write based on 'attr'
        if attr is not None:
            attr_name, attr_get, attr_set = attr
            
            if attr_get is None:
                self._read_data = lambda s,x: s.__setattr__(attr_name, x)
            else:
                self._read_data = lambda s,x: s.__setattr__(attr_name, attr_get(x))

            if attr_set is None:
                self._write_data = lambda s: str_no_none(s.__dict__[attr_name])
            else:
                self._write_data = lambda s: attr_set(s.__dict__[attr_name])

        
        # init tag lookup
        self._tags = {}
        for tag in tags:
            self._tags[tag.name] = tag

        # set of initialized tags
        self._init_tags = __builtins__["set"]()


    #===========================================
    # reading

    def init(self):
        """Initialize the a tag before its first use"""
        self._init_tags.clear()
        

    def set_object(self, obj):
        self._object = obj


    def new_tag(self, name):
        """Create new child tag"""
        
        tag = self._tags.get(name, None)
        if tag:
            # initialize new tag
            tag.set_object(self._object)

            if tag not in self._init_tags:
                tag.init()
                self._init_tags.add(tag)                
        return tag

    def start_tag(self):
        """Start tag callback"""
        self._data = []

    def queue_data(self, data):
        """Content data callback"""
        
        if self._read_data:
            self._data.append(data)

    def end_tag(self):
        """End tag callback"""

        # read queued data if read function is supplied
        if self._read_data:
            data = "".join(self._data)
            self._data = []
            
            try:
                self._read_data(self._object, data)
            except Exception, e:
                raise XmlError("Error parsing tag '%s': %s" % (self.name,
                                                               str(e)))
    
    def add(self, tag):
        """Add a tag child to this tag"""
        
        self._tag_list.append(tag)
        self._tags[tag.name] = tag


    #===================
    # writing

    def write(self, obj, out):
        """Write tag to output stream"""
        
        # write openning        
        if self.name != "":
            out.write("<%s>" % self.name)
        
        if len(self._tags) > 0:
            out.write("\n")
            for child_tag in self._tag_list:
                child_tag.write(obj, out)
        elif self._write_data:
            text = self._write_data(obj)
            if not isinstance(text, basestring):
                raise XmlError("bad text (%s,%s): %s" %
                               (self.name, str(self._object),
                                str(type(text))))
            out.write(escape(text))
        
        if self.name != "":
            out.write("</%s>\n" % self.name)
        
    
# TODO: remove get?

class TagMany (Tag):
    def __init__(self, name, iterfunc, get=None, set=None,
                 before=None,
                 after=None,
                 tags=[]):
        Tag.__init__(self, name,
                     get=None,
                     set=set,
                     tags=tags)
        
        self._iterfunc = iterfunc
        self._read_item = get
        self._write_item = set
        self._beforefunc = before
        self._afterfunc = after
        self._index = 0


    #=============================
    # reading

    def init(self):
        """Initialize the a tag before its first use"""
        self._init_tags.clear()
        self._index = 0
        

    def new_tag(self, name):
        """Create new child tag"""
        tag = self._tags.get(name, None)
        if tag:
            # initialize new tag
            tag.set_object((self._object, self._index))

            if tag not in self._init_tags:
                tag.init()
                self._init_tags.add(tag)                 
        return tag

    def start_tag(self):
        """Start tag callback"""
        self._data = []
        if self._beforefunc:
            self._beforefunc((self._object, self._index))

    def queue_data(self, data):
        """Content data callback"""
        if self._read_item:
            self._data.append(data)

    def end_tag(self):
        """End tag callback"""
        
        if self._read_item:
            data = "".join(self._data)
            self._data = []
            
            #try:
            if 1:
                if self._read_item is not None:
                    self._read_item((self._object, self._index), data)
            #except Exception, e:
            #    raise XmlError("Error parsing tag '%s': %s" % (self.name,
            #                                                   str(e)))
        
        if self._afterfunc:
            self._afterfunc((self._object, self._index))
        self._index += 1


    #=====================
    # writing
    
    def write(self, obj, out):
        # write opening
        if len(self._tags) == 0:
            assert self._write_item is not None
                
            for i in self._iterfunc(obj):
                out.write("<%s>%s</%s>\n" % (self.name,
                                             escape(self._write_item((obj, i))),
                                             self.name))
        else:
            for i in self._iterfunc(obj):
                out.write("<%s>\n" % self.name)
                for child_tag in self._tag_list:
                    child_tag.write((obj, i), out)
                out.write("</%s>\n" % self.name)

'''
# TODO: remove get?

class TagList (TagMany):
    """A specialization of TagMany to work with reading and writing lists"""

    def __init__(self, name, lst, get=None, set=None, before=None, after=None,
                 tags=[]):
        TagMany.__init__(self, name, self._iter,
                         get=get, set=set,
                         before=before, after=after, tags=tags)

        self._list = lst

        def new_tag(self, name):
            tag = self._tags.get(name, None)
            if tag:
                tag.set_object(self._list)
            return tag
'''            
        

class XmlObject (object):
    """Represents an object <--> XML document binding"""
    
    def __init__(self, *tags):
        self._object = None
        self._root_tag = Tag("", tags=tags)
        self._current_tags = [self._root_tag]
        
    
    def __start_element(self, name, attrs):
        """Start tag callback"""
        
        if len(self._current_tags) > 0:
            last_tag = self._current_tags[-1]
            if last_tag:
                new_tag = last_tag.new_tag(name)
                self._current_tags.append(new_tag)
                if new_tag:                
                    new_tag.start_tag()
            
        
    def __end_element(self, name):
        """End tag callback"""
        
        if len(self._current_tags) > 0:
            last_tag = self._current_tags.pop()
            if last_tag:
                if last_tag.name == name:
                    last_tag.end_tag()
                else:
                    raise XmlError("Malformed XML")            

                
        
    def __char_data(self, data):
        """read character data and give it to current tag"""

        if len(self._current_tags) > 0:
            tag = self._current_tags[-1]
            if tag:
                tag.queue_data(data)
            
            
    
    def read(self, obj, filename):
        """Read XML from 'filename' and store data into object 'obj'"""
        
        if isinstance(filename, basestring):
            infile = open(filename, "r")
        else:
            infile = filename
        self._object = obj
        self._root_tag.set_object(self._object)
        self._current_tags = [self._root_tag]
        self._root_tag.init()
        
        parser = xml.parsers.expat.ParserCreate()
        parser.StartElementHandler = self.__start_element
        parser.EndElementHandler = self.__end_element
        parser.CharacterDataHandler = self.__char_data

        try:
            parser.ParseFile(infile)
        except xml.parsers.expat.ExpatError, e:
            raise XmlError("Error reading file '%s': %s" % (filename, str(e)))

        if len(self._current_tags) > 1:
            print [x.name for x in self._current_tags]
            raise XmlError("Incomplete file '%s'" % filename)
        
        infile.close()

            
    def write(self, obj, filename):
        """Write object 'obj' to file 'filename'"""
        
        if isinstance(filename, basestring):
            #out = codecs.open(filename, "w", "utf-8")
            out = safefile.open(filename, "w", codec="utf-8")            
            need_close = True
        else:
            out = filename
            need_close = False
        
        out.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>")
        self._root_tag.write(obj, out)
        out.write("\n")
        if need_close:
            out.close()
        
        


if __name__ == "__main__":
    import StringIO

    parser = XmlObject(
        Tag("notebook", tags=[
            Tag("window_size",
                attr=("window_size",
                      lambda x: tuple(map(int, x.split(","))),
                      lambda x: "%d,%d" % x)),
            Tag("window_pos",
                attr=("window_pos",
                      lambda x: tuple(map(int, x.split(","))),
                      lambda x: "%d,%d" % x)),
            Tag("vsash_pos",
                attr=("vhash_pos", int, str)),
            Tag("hsash_pos",
                attr=("hsash_pos", int, str)),
            Tag("external_apps", tags=[
                TagMany("app",
                        iterfunc=lambda s: range(len(s.apps)),
                        get=lambda (s,i), x: s.apps.append(x),
                        set=lambda (s,i): s.apps[i])]),
            Tag("external_apps2", tags=[
                TagMany("app",
                        iterfunc=lambda s: range(len(s.apps2)),
                        before=lambda (s,i): s.apps2.append([None, None]),
                        tags=[Tag("name",
                                  get=lambda (s,i),x: s.apps2[i].__setitem__(0, x),
                                  set=lambda (s,i): s.apps2[i][0]),
                              Tag("prog",
                                  get=lambda (s,i),x: s.apps2[i].__setitem__(1,x),
                                  set=lambda (s,i): s.apps2[i][1])
                        ])
            ]),
        ]))

    class Pref (object):
        def __init__(self):
            self.window_size = (0, 0)
            self.window_pos = (0, 0)
            self.vsash_pos = 0
            self.hsash_pos = 0
            self.apps = []
            self.apps2 = []
            
        def read(self, filename):
            parser.read(self, filename)
           
        def write(self, filename):
            parser.write(self, filename)
    
    #from rasmus import util
    
    #util.tic("run")

    infile = StringIO.StringIO("""<?xml version="1.0" encoding="UTF-8"?>
       <notebook>
       <window_size>1053,905</window_size>
<window_pos>0,0</window_pos>
<vsash_pos>0</vsash_pos>
<hsash_pos>250</hsash_pos>
<external_apps>
<app>web_browser</app>
<app>image_editor</app>
</external_apps>
<external_apps2>
<app><name>web_browser</name><prog>firefox</prog></app>
<app><name>image_editor</name><prog>gimp</prog></app>
</external_apps2>
       </notebook>
    """)
    
    for i in xrange(1):#0000):
        pref = Pref()
        pref.read(infile)
        pref.write(sys.stdout)
    
    #util.toc()
    




'''
def get_dom_children(node):
    """Convenience function for iterating the children of a DOM object"""
    child = node.firstChild
    while child:
        yield child
        child = child.nextSibling
'''




"""

    KeepNote
    Backward compatiability for configuration information

"""

#
#  KeepNote
#  Copyright (c) 2008-2009 Matt Rasmussen
#  Author: Matt Rasmussen <rasmus@mit.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#

import os
import shutil
from xml.etree import ElementTree

import keepnote
from keepnote import FS_ENCODING
from keepnote import xdg

OLD_USER_PREF_DIR = u"takenote"
OLD_USER_PREF_FILE = u"takenote.xml"
OLD_XDG_USER_EXTENSIONS_DIR = u"takenote/extensions"
OLD_XDG_USER_EXTENSIONS_DATA_DIR = u"takenote/extensions_data"


USER_PREF_DIR = u"keepnote"
USER_PREF_FILE = u"keepnote.xml"
USER_EXTENSIONS_DIR = u"extensions"
USER_EXTENSIONS_DATA_DIR = u"extensions_data"

XDG_USER_EXTENSIONS_DIR = u"keepnote/extensions"
XDG_USER_EXTENSIONS_DATA_DIR = u"keepnote/extensions_data"



def get_old_pref_dir1(home):
    """
    Returns old preference directory (type 1)
    $HOME/.takenote
    """
    return os.path.join(home, "." + OLD_USER_PREF_DIR)


def get_old_pref_dir2(home):
    """
    Returns old preference directory (type 2)
    $HOME/.config/takenote
    """
    return os.path.join(home, "config", OLD_USER_PREF_DIR)



def get_new_pref_dir(home):
    """
    Returns old preference directory (type 2)
    $HOME/.config/takenote
    """
    return os.path.join(home, "config", USER_PREF_DIR)


def get_home():
    """Return HOME directory"""
    home = keepnote.ensure_unicode(os.getenv(u"HOME"), FS_ENCODING)
    if home is None:
        raise EnvError("HOME environment variable must be specified")
    return home


def get_old_user_pref_dir(home=None):
    """Returns the directory of the application preference file"""
    
    p = keepnote.get_platform()
    if p == "unix" or p == "darwin":
        
        if home is None:
            home = get_home()
        old_dir = get_old_pref_dir1(home)

        if os.path.exists(old_dir):
            return old_dir
        else:
            return xdg.get_config_file(OLD_USER_PREF_DIR, default=True)

    elif p == "windows":
        appdata = keepnote.ensure_unicode(os.getenv(u"APPDATA"), FS_ENCODING)
        if appdata is None:
            raise keepnote.EnvError("APPDATA environment variable must be specified")
        return os.path.join(appdata, OLD_USER_PREF_DIR)

    else:
        raise Exception("unknown platform '%s'" % p)


def get_new_user_pref_dir(home=None):
    """Returns the directory of the application preference file"""
    
    p = keepnote.get_platform()
    if p == "unix" or p == "darwin":
        
        if home is None:
            home = get_home()
        return xdg.get_config_file(USER_PREF_DIR, default=True)

    elif p == "windows":
        appdata = keepnote.ensure_unicode(os.getenv(u"APPDATA"), FS_ENCODING)
        if appdata is None:
            raise keepnote.EnvError("APPDATA environment variable must be specified")
        return os.path.join(appdata, USER_PREF_DIR)

    else:
        raise Exception("unknown platform '%s'" % p)


def upgrade_user_pref_dir(old_user_pref_dir, new_user_pref_dir):
    """Moves preference data from old location to new one"""

    import sys
    
    # move user preference directory
    shutil.copytree(old_user_pref_dir, new_user_pref_dir)

    # rename takenote.xml to keepnote.xml
    newfile = os.path.join(new_user_pref_dir, USER_PREF_FILE)
    os.rename(os.path.join(new_user_pref_dir, OLD_USER_PREF_FILE), newfile)
    
    # rename root xml tag
    tree = ElementTree.ElementTree(file=newfile)
    elm = tree.getroot()
    elm.tag = "keepnote"
    tree.write(newfile, encoding="UTF-8")

    # move over data files from .local/share/takenote
    if keepnote.get_platform() in ("unix", "darwin"):
        datadir = os.path.join(get_home(), ".local", "share", "takenote")
        
        old_ext_dir = os.path.join(datadir, "extensions")
        new_ext_dir = os.path.join(new_user_pref_dir, "extensions")    
        if not os.path.exists(new_ext_dir) and os.path.exists(old_ext_dir):
            shutil.copytree(old_ext_dir, new_ext_dir)

        old_ext_dir = os.path.join(datadir, "extensions_data")
        new_ext_dir = os.path.join(new_user_pref_dir, "extensions_data")    
        if not os.path.exists(new_ext_dir) and os.path.exists(old_ext_dir):
            shutil.copytree(old_ext_dir, new_ext_dir)
            
        


def check_old_user_pref_dir(home=None):
    """Upgrades user preference directory if it exists in an old format"""

    old_pref_dir = get_old_user_pref_dir(home)
    new_pref_dir = get_new_user_pref_dir(home)
    if not os.path.exists(new_pref_dir) and os.path.exists(old_pref_dir):
        upgrade_user_pref_dir(old_pref_dir, new_pref_dir)
        





"""

    KeepNote

    timestamp module

"""

#
#  KeepNote
#  Copyright (c) 2008-2009 Matt Rasmussen
#  Author: Matt Rasmussen <rasmus@mit.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#

import locale
import sys
import time


# determine UNIX Epoc (which should be 0, unless the current platform has a 
# different definition of epoc)
# Use the epoc date + 1 month (SEC_OFFSET) in order to prevent underflow in date due to user's timezone
SEC_OFFSET = 3600 * 24 * 31
EPOC = time.mktime((1970, 2, 1, 0, 0, 0, 3, 1, 0)) - time.timezone - SEC_OFFSET


"""

0  	tm_year  	(for example, 1993)
1 	tm_mon 	range [1,12]
2 	tm_mday 	range [1,31]
3 	tm_hour 	range [0,23]
4 	tm_min 	range [0,59]
5 	tm_sec 	range [0,61]; see (1) in strftime() description
6 	tm_wday 	range [0,6], Monday is 0
7 	tm_yday 	range [1,366]
8 	tm_isdst 	0, 1 or -1; see below

"""

TM_YEAR, \
TM_MON, \
TM_MDAY, \
TM_HOUR, \
TM_MIN, \
TM_SEC, \
TM_WDAY, \
TM_YDAY, \
TM_ISDST = range(9)



"""
%a  	Locale's abbreviated weekday name.  	
%A 	Locale's full weekday name. 	
%b 	Locale's abbreviated month name. 	
%B 	Locale's full month name. 	
%c 	Locale's appropriate date and time representation. 	
%d 	Day of the month as a decimal number [01,31]. 	
%H 	Hour (24-hour clock) as a decimal number [00,23]. 	
%I 	Hour (12-hour clock) as a decimal number [01,12]. 	
%j 	Day of the year as a decimal number [001,366]. 	
%m 	Month as a decimal number [01,12]. 	
%M 	Minute as a decimal number [00,59]. 	
%p 	Locale's equivalent of either AM or PM. 	(1)
%S 	Second as a decimal number [00,61]. 	(2)
%U 	Week number of the year (Sunday as the first day of the week) as a decimal number [00,53]. All days in a new year preceding the first Sunday are considered to be in week 0. 	(3)
%w 	Weekday as a decimal number [0(Sunday),6]. 	
%W 	Week number of the year (Monday as the first day of the week) as a decimal number [00,53]. All days in a new year preceding the first Monday are considered to be in week 0. 	(3)
%x 	Locale's appropriate date representation. 	
%X 	Locale's appropriate time representation. 	
%y 	Year without century as a decimal number [00,99]. 	
%Y 	Year with century as a decimal number. 	
%Z 	Time zone name (no characters if no time zone exists). 	
%% 	A literal "%" character.
"""

DEFAULT_TIMESTAMP_FORMATS = {
    "same_day": u"%I:%M %p",
    "same_month": u"%a, %d %I:%M %p",
    "same_year": u"%a, %b %d %I:%M %p",
    "diff_year": u"%a, %b %d, %Y"
}



def get_timestamp():
    """Returns the current timestamp"""
    return int(time.time() - EPOC)

def get_localtime():
    """Returns the local time"""
    return time.localtime()

def get_str_timestamp(timestamp, current=None,
                      formats=DEFAULT_TIMESTAMP_FORMATS):
    """
    Get a string representation of a time stamp
    
    The string will be abbreviated according to the current time.
    """

    # NOTE: I have written this function to allow unicode formats.
    # The encode/decode functions should allow most unicode formats to
    # to be processed by strftime.  However, a '%' may occur inside a 
    # multibyte character.  This is a hack until python issue
    # http://bugs.python.org/issue2782 is resolved.

    encoding = locale.getdefaultlocale()[1]
    if encoding is None:
        encoding = "utf-8"

    if formats is None:
        formats = DEFAULT_TIMESTAMP_FORMATS

    try:
        if current is None:
            current = get_localtime()
        local = time.localtime(timestamp + EPOC)
    
        if local[TM_YEAR] == current[TM_YEAR]:
            if local[TM_MON] == current[TM_MON]:
                if local[TM_MDAY] == current[TM_MDAY]:
                    return time.strftime(formats["same_day"].encode(encoding), 
                                         local).decode(encoding)
                else:
                    return time.strftime(formats["same_month"].encode(encoding),
                                         local).decode(encoding)
            else:
                return time.strftime(formats["same_year"].encode(encoding), 
                                     local).decode(encoding)
        else:
            return time.strftime(formats["diff_year"].encode(encoding), 
                                 local).decode(encoding)
    except:
        return u"[formatting error]"






"""
    KeepNote Extension 
    backup_tar

    Tar file notebook backup
"""

#
#  KeepNote
#  Copyright (c) 2008-2009 Matt Rasmussen
#  Author: Matt Rasmussen <rasmus@mit.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#


# python imports
import os
import sys

# keepnote imports
import keepnote
from keepnote import AppCommand
from keepnote.gui import extension


class Extension (extension.Extension):
    
    version = (1, 0)
    name = "Basic Commands"
    author = "Matt Rasmussen <rasmus@mit.edu>"
    description = "Adds basic command line options to KeepNote"


    def __init__(self, app):
        """Initialize extension"""
        
        extension.Extension.__init__(self, app)
        self.app = app
        self.enabled.add(self.on_enabled)

        self.commands = [
            AppCommand("focus", lambda app, args: app.focus_windows(),
                       help="focus all open windows"),
            AppCommand("screenshot", self.on_screenshot,
                       help="insert a new screenshot"),
            AppCommand("install", self.on_install_extension,
                       metavar="FILENAME",
                       help="install a new extension"),
            AppCommand("uninstall", self.on_uninstall_extension,
                       metavar="EXTENSION_NAME",
                       help="uninstall an extension"),
            AppCommand("quit", lambda app, args: app.quit(),
                       help="close all KeepNote windows"),
            ]


    def get_depends(self):
        return [("keepnote", ">=", (0, 6, 2))]


    def on_enabled(self, enabled):
        
        if enabled:
            for command in self.commands:
                if self.app.get_command(command.name):
                    continue

                try:
                    self.app.add_command(command)
                except Exception, e:
                    self.app.error("Could not add command '%s'" % command.name,
                                   e, sys.exc_info()[2])

        else:
            for command in self.commands:
                self.app.remove_command(command.name)


    #====================================================
    # commands

    def on_uninstall_extension(self, app, args):
        
        for extname in args[1:]:
            ext = app.get_extension(extname)
            if ext is None:
                app.error("unknown extension '%s'" % extname)
            else:
                app.uninstall_extension(ext)


    def on_install_extension(self, app, args):
        
        for filename in args[1:]:
            app.install_extension(filename)


    def on_screenshot(self, app, args):
        window = app.get_current_window()
        if window:
            window.get_viewer().editor.on_screenshot()
        



"""

    KeepNote
    utilities

"""

#
#  KeepNote
#  Copyright (c) 2008-2009 Matt Rasmussen
#  Author: Matt Rasmussen <rasmus@mit.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
#



class PushIter (object):
    """Wrap an iterator in another iterator that allows one to push new
       items onto the front of the iteration stream"""
    
    def __init__(self, it):
        self._it = iter(it)
        self._queue = []

    def __iter__(self):
        return self
        
    def next(self):
        if len(self._queue) > 0:
            return self._queue.pop()
        else:
            return self._it.next()

    def push(self, item):
        """Push a new item onto the front of the iteration stream"""
        self._queue.append(item)




def compose2(f, g):
    """
    Compose two functions into one

    compose2(f, g)(x) <==> f(g(x))
    """
    return lambda *args, **kargs: f(g(*args, **kargs))
    

def compose(*funcs):
    """Composes two or more functions into one function
    
       example:
       compose(f,g)(x) <==> f(g(x))
    """

    funcs = reversed(funcs)
    f = funcs.next()
    for g in funcs:
        f = compose2(g, f)
    return f

